from typing import Any, Dict, Tuple

from pytest import raises
from tune.exceptions import TuneCompileError
from tune.concepts.flow import Trial, TrialReport
from tune.noniterative.convert import noniterative_objective, to_noniterative_objective
from tune.noniterative.objective import NonIterativeObjectiveFunc


def test_to_noniterative_objective():
    def f1(a, b: int) -> float:
        return a - b

    func = to_noniterative_objective(f1)
    assert func.min_better
    trial = Trial("abc", dict(b=20, a=10), dict(c=3))
    report = func.run(trial)
    assert report.trial is trial
    assert report.metric == -10
    assert report.params == trial.params
    assert report.metadata == {}

    func = to_noniterative_objective("f1", min_better=False)
    assert not func.min_better
    trial = Trial("abc", dict(b=20, a=10), dict(c=3))
    report = func.run(trial)
    assert report.trial is trial
    assert report.metric == -10
    assert report.params == trial.params
    assert report.metadata == {}

    assert -1 == func(1, 2)

    def f2(a, b: int) -> Tuple[float, Dict[str, Any]]:
        return a - b, dict(c=5)

    func = to_noniterative_objective(f2)
    trial = Trial("abc", dict(b=20, a=10), dict(c=3))
    report = func.run(trial)
    assert report.trial is trial
    assert report.metric == -10
    assert report.params == trial.params
    assert report.metadata == dict(c=5)

    def f3(t: Trial) -> TrialReport:
        return TrialReport(
            t, t.params["a"] - t.params["b"], params=dict(a=1), metadata=dict(c=6)
        )

    func = to_noniterative_objective(f3)
    trial = Trial("abc", dict(b=20, a=10), dict(c=3))
    report = func.run(trial)
    assert report.trial is trial
    assert report.metric == -10
    assert report.params == dict(a=1)
    assert report.metadata == dict(c=6)

    class F4(NonIterativeObjectiveFunc):
        def run(self, t: Trial) -> TrialReport:
            return TrialReport(
                t, t.params["a"] - t.params["b"], params=dict(a=1), metadata=dict(c=6)
            )

    f4 = F4()
    f4_ = to_noniterative_objective(f4)
    assert isinstance(f4_, F4)
    assert f4 is not f4_

    raises(TuneCompileError, lambda: to_noniterative_objective("abc"))


def test_noniterative_objective():
    @noniterative_objective
    def f1(a, b: int) -> float:
        return a - b

    assert isinstance(f1, NonIterativeObjectiveFunc)
    trial = Trial("abc", dict(b=20, a=10), dict(c=3))
    report = f1.run(trial)
    assert report.trial is trial
    assert report.metric == -10
    assert report.params == trial.params
    assert report.metadata == {}

    func = to_noniterative_objective("f1")
    report = func.run(trial)
    assert report.trial is trial
    assert report.metric == -10
    assert report.params == trial.params
    assert report.metadata == {}

    @noniterative_objective(min_better=False)
    def f2(a, b: int) -> float:
        return a - b

    assert isinstance(f2, NonIterativeObjectiveFunc)
    assert not f2.min_better



# flake8: noqa

from tune.concepts.space.parameters import (
    Choice,
    Grid,
    NormalRand,
    NormalRandInt,
    Rand,
    RandInt,
    StochasticExpression,
    decode_params,
    encode_params,
)
from tune.concepts.space.spaces import HorizontalSpace, Space, VerticalSpace



from typing import Any, Dict, Iterable

from fugue import FugueWorkflow
from tune.concepts.dataset import TuneDatasetBuilder
from tune.iterative.objective import IterativeObjectiveFunc
from tune import optimize_by_sha, optimize_by_hyperband
from tune.concepts.space import Grid, Space
from tune.concepts.flow import TrialReport
from tune.constants import TUNE_REPORT_METRIC


class F(IterativeObjectiveFunc):
    def __init__(self):
        super().__init__()
        self._it = 0
        self._all = [
            [9, 3, 1, 1],
            [8, 6, 5, 5],
            [8, 5, 4, 3],
            [7, 4, 3, 4],
        ]

    def save_checkpoint(self, fs):
        fs.writetext("x", str(self._it))

    def load_checkpoint(self, fs):
        self._it = int(fs.readtext("x"))

    def run_single_iteration(self):
        trial = self.current_trial
        metric = self._all[trial.params["a"]][self._it]
        self._it += 1
        return TrialReport(trial, metric=metric)

    def copy(self):
        return F()


def test_sha(tmpdir):
    def assert_metric(df: Iterable[Dict[str, Any]], metric: float, ct: int) -> None:
        n = 0
        for row in df:
            assert row[TUNE_REPORT_METRIC] == metric
            n += 1
        assert n == ct

    space = Space(a=Grid(0, 1, 2, 3))
    dag = FugueWorkflow()
    dataset = TuneDatasetBuilder(space, str(tmpdir)).build(dag)
    obj = F()
    res = optimize_by_sha(
        obj,
        dataset,
        plan=[[1.0, 3], [1.0, 2], [1.0, 1], [1.0, 1]],
        checkpoint_path=str(tmpdir),
    )
    res.result().output(assert_metric, dict(metric=4.0, ct=1))

    res = optimize_by_sha(
        obj,
        dataset,
        plan=[[2.0, 2], [1.0, 1], [1.0, 1]],
        checkpoint_path=str(tmpdir),
    )
    res.result().output(assert_metric, dict(metric=1.0, ct=1))
    dag.run()


def test_hyperband(tmpdir):
    def assert_metric(df: Iterable[Dict[str, Any]], metric: float, ct: int) -> None:
        n = 0
        for row in df:
            if metric > 0:
                assert row[TUNE_REPORT_METRIC] == metric
            n += 1
        assert n == ct

    space = Space(a=Grid(0, 1, 2, 3))
    dag = FugueWorkflow()
    dataset = TuneDatasetBuilder(space, str(tmpdir)).build(dag)
    obj = F()
    res = optimize_by_hyperband(
        obj,
        dataset,
        plans=[
            [[1.0, 3], [1.0, 2], [1.0, 1], [1.0, 1]],
            [[2.0, 2], [1.0, 1], [1.0, 1]],
        ],
        checkpoint_path=str(tmpdir),
    )
    res.result().output(assert_metric, dict(metric=0.0, ct=2))
    res.result(1).output(assert_metric, dict(metric=1.0, ct=1))
    dag.run()



from typing import Iterable, Any, List


def assert_close(seq1: Iterable[Any], seq2: Iterable[Any], error: float = 1e-5):
    def dedup(seq: List[Any]) -> Iterable[Any]:
        last: Any = None
        for x in sorted(seq):
            if last is None:
                last = x
                yield x
            elif abs(x - last) < error:
                continue
            else:
                last = x
                yield x

    _s1, _s2 = list(dedup(seq1)), list(dedup(seq2))  # type: ignore
    assert len(_s1) == len(_s2), f"{_s1},{_s2}"
    for x, y in zip(_s1, _s2):
        assert abs(x - y) < error, f"{_s1},{_s2}"



from fugue_dask import DaskExecutionEngine
from sklearn.datasets import load_diabetes
from sklearn.linear_model import Lasso, LinearRegression
from tune import TUNE_OBJECT_FACTORY, Grid, Rand
from tune_hyperopt.optimizer import HyperoptLocalOptimizer
from tune_sklearn import sk_space, suggest_sk_models, suggest_sk_models_by_cv

# from fugue_spark import SparkExecutionEngine


def test_suggest(tmpdir):
    TUNE_OBJECT_FACTORY.set_temp_path(str(tmpdir))

    data = load_diabetes(as_frame=True)
    alldata = data["data"]
    alldata["label"] = data["target"]
    alldata = alldata.sample(frac=1, random_state=0)
    test = alldata.iloc[:20]
    train = alldata.iloc[20:]

    space1 = sk_space(LinearRegression, fit_intercept=Grid(True, False))
    space2 = sk_space(
        Lasso, fit_intercept=Grid(True, False), alpha=Rand(0.5, 1)
    ).sample(3, 0)
    space3 = sk_space(Lasso, fit_intercept=Grid(True, False), alpha=Rand(0.5, 1))
    result = suggest_sk_models(
        space1 + space3,
        train,
        test,
        "neg_mean_absolute_error",
        top_n=0,
        distributed=False,
        local_optimizer=HyperoptLocalOptimizer(max_iter=10, seed=0),
    )
    # print([x.jsondict for x in result])
    assert 4 == len(result)
    assert 50 > result[0].sort_metric

    result = suggest_sk_models(
        space1 + space2,
        train,
        test,
        "neg_mean_absolute_error",
        top_n=0,
        partition_keys=["sex"],
        temp_path=str(tmpdir),
        save_model=True,
    )
    assert 16 == len(result)
    assert 50 > result[0].sort_metric

    result = suggest_sk_models(
        space1 + space2,
        train,
        test,
        "neg_mean_absolute_error",
        top_n=1,
        partition_keys=["sex"],
        execution_engine=DaskExecutionEngine,
    )
    assert 2 == len(result)
    assert 50 > result[0].sort_metric


def test_suggest_cv(tmpdir):
    TUNE_OBJECT_FACTORY.set_temp_path(str(tmpdir))

    data = load_diabetes(as_frame=True)
    train = data["data"]
    train["label"] = data["target"]

    space1 = sk_space(LinearRegression, fit_intercept=Grid(True, False))
    space2 = sk_space(
        Lasso, fit_intercept=Grid(True, False), alpha=Rand(0.5, 1)
    ).sample(3, 0)
    space3 = sk_space(Lasso, fit_intercept=Grid(True, False), alpha=Rand(0.5, 1))
    result = suggest_sk_models_by_cv(
        space1 + space3,
        train,
        "neg_mean_absolute_error",
        top_n=0,
        distributed=False,
        local_optimizer=HyperoptLocalOptimizer(max_iter=10, seed=0),
    )
    # print([x.jsondict for x in result])
    assert 4 == len(result)
    assert 50 > result[0].sort_metric

    result = suggest_sk_models_by_cv(
        space1 + space2,
        train,
        "neg_mean_absolute_error",
        top_n=0,
        partition_keys=["sex"],
        temp_path=str(tmpdir),
        save_model=True,
    )
    assert 16 == len(result)
    assert 50 > result[0].sort_metric

    result = suggest_sk_models_by_cv(
        space1 + space2,
        train,
        "neg_mean_absolute_error",
        top_n=1,
        partition_keys=["sex"],
        execution_engine=DaskExecutionEngine,
    )
    assert 2 == len(result)
    assert 50 > result[0].sort_metric

    # TODO: why the following unit test is so slow

    # result = suggest_sk_models_by_cv(
    #     space1 + space2,
    #     train,
    #     "neg_mean_absolute_error",
    #     top_n=1,
    #     partition_keys=["sex"],
    #     execution_engine=SparkExecutionEngine,
    #     execution_engine_conf={
    #         "spark.sql.shuffle.partitions": 4,
    #         "spark.default.parallelism": 4,
    #         "spark.executor.cores": 4,
    #         "spark.sql.adaptive.enabled": "false",
    #     },
    # )
    # assert 2 == len(result)
    # assert 50 > result[0].sort_metric



import pandas as pd
from tests.tune_tensorflow.mock import MockSpec
from tune import Trial, validate_iterative_objective

from tune_tensorflow import keras_space
from tune_tensorflow.objective import KerasObjective
from tune_tensorflow.utils import _TYPE_DICT


def test_spec():
    spec = MockSpec(dict(l1=16, l2=16), {})
    metric = spec.compute_sort_metric(epochs=10)
    assert metric < 15

    spec = MockSpec(dict(l1=16, l2=16), {"x": pd.DataFrame([[0]], columns=["a"])})
    metric = spec.compute_sort_metric(epochs=10)
    assert metric < 15
    assert isinstance(spec.dfs["x"], pd.DataFrame)


def test_objective():
    def validate(reports):
        assert reports[-1].metric < 15

    space = keras_space(MockSpec, l1=16, l2=16)
    obj = KerasObjective(_TYPE_DICT)

    for cont in [True, False]:
        validate_iterative_objective(
            obj,
            Trial("a", params=list(space)[0]),
            budgets=[3, 3, 4],
            continuous=cont,
            validator=validate,
        )



# flake8: noqa

from tune_version import __version__

from tune.api.factory import TUNE_OBJECT_FACTORY
from tune.api.optimize import (
    optimize_by_continuous_asha,
    optimize_by_hyperband,
    optimize_by_sha,
    optimize_noniterative,
)
from tune.api.suggest import (
    suggest_by_continuous_asha,
    suggest_by_hyperband,
    suggest_by_sha,
    suggest_for_noniterative_objective,
)
from tune.concepts.checkpoint import Checkpoint
from tune.concepts.dataset import StudyResult, TuneDataset, TuneDatasetBuilder
from tune.concepts.flow import (
    Monitor,
    Trial,
    TrialDecision,
    TrialReport,
    TrialReportLogger,
)
from tune.concepts.space import Choice, Grid, Rand, RandInt, Space, StochasticExpression
from tune.iterative.objective import (
    IterativeObjectiveFunc,
    validate_iterative_objective,
)
from tune.noniterative.convert import noniterative_objective, to_noniterative_objective
from tune.noniterative.objective import (
    NonIterativeObjectiveFunc,
    NonIterativeObjectiveLocalOptimizer,
    validate_noniterative_objective,
)
from tune.noniterative.stopper import (
    NonIterativeStopper,
    n_samples,
    n_updates,
    no_update_period,
    small_improvement,
)



import json
from typing import Any, List, Optional, Tuple

from fugue import FugueWorkflow
from fugue.exceptions import FugueDataFrameError
from triad import assert_or_throw
from tune.api.factory import TUNE_OBJECT_FACTORY
from tune.api.optimize import (
    optimize_by_continuous_asha,
    optimize_by_hyperband,
    optimize_by_sha,
    optimize_noniterative,
)
from tune.concepts.flow import TrialReport
from tune.concepts.space import Space
from tune.constants import TUNE_DATASET_DF_DEFAULT_NAME, TUNE_REPORT, TUNE_REPORT_METRIC
from tune.exceptions import TuneCompileError


def suggest_for_noniterative_objective(
    objective: Any,
    space: Space,
    df: Any = None,
    df_name: str = TUNE_DATASET_DF_DEFAULT_NAME,
    temp_path: str = "",
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    local_optimizer: Any = None,
    monitor: Any = None,
    stopper: Any = None,
    stop_check_interval: Any = None,
    distributed: Optional[bool] = None,
    shuffle_candidates: bool = True,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    dag = FugueWorkflow()
    dataset = TUNE_OBJECT_FACTORY.make_dataset(
        dag,
        space,
        df=df,
        df_name=df_name,
        partition_keys=partition_keys,
        temp_path=temp_path,
        shuffle=shuffle_candidates,
    )
    study = optimize_noniterative(
        objective=objective,
        dataset=dataset,
        optimizer=local_optimizer,
        distributed=distributed,
        monitor=monitor,
        stopper=stopper,
        stop_check_interval=stop_check_interval,
    )
    study.result(top_n).yield_dataframe_as("result")

    return _run(
        dag=dag,
        execution_engine=execution_engine,
        execution_engine_conf=execution_engine_conf,
    )


def suggest_by_sha(
    objective: Any,
    space: Space,
    plan: List[Tuple[float, int]],
    train_df: Any = None,
    temp_path: str = "",
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    monitor: Any = None,
    distributed: Optional[bool] = None,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    assert_or_throw(
        not space.has_random_parameter,
        TuneCompileError(
            "space can't contain random parameters, "
            "use sample method before calling this function"
        ),
    )
    dag = FugueWorkflow()
    dataset = TUNE_OBJECT_FACTORY.make_dataset(
        dag,
        space,
        df=train_df,
        partition_keys=partition_keys,
        temp_path=temp_path,
    )
    study = optimize_by_sha(
        objective=objective,
        dataset=dataset,
        plan=plan,
        checkpoint_path=temp_path,
        distributed=distributed,
        monitor=monitor,
    )
    study.result(top_n).yield_dataframe_as("result")

    return _run(
        dag=dag,
        execution_engine=execution_engine,
        execution_engine_conf=execution_engine_conf,
    )


def suggest_by_hyperband(
    objective: Any,
    space: Space,
    plans: List[List[Tuple[float, int]]],
    train_df: Any = None,
    temp_path: str = "",
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    monitor: Any = None,
    distributed: Optional[bool] = None,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    assert_or_throw(
        not space.has_random_parameter,
        TuneCompileError(
            "space can't contain random parameters, "
            "use sample method before calling this function"
        ),
    )
    dag = FugueWorkflow()
    dataset = TUNE_OBJECT_FACTORY.make_dataset(
        dag,
        space,
        df=train_df,
        partition_keys=partition_keys,
        temp_path=temp_path,
    )
    study = optimize_by_hyperband(
        objective=objective,
        dataset=dataset,
        plans=plans,
        checkpoint_path=temp_path,
        distributed=distributed,
        monitor=monitor,
    )
    study.result(top_n).yield_dataframe_as("result")

    return _run(
        dag=dag,
        execution_engine=execution_engine,
        execution_engine_conf=execution_engine_conf,
    )


def suggest_by_continuous_asha(
    objective: Any,
    space: Space,
    plan: List[Tuple[float, int]],
    train_df: Any = None,
    temp_path: str = "",
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    monitor: Any = None,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    assert_or_throw(
        not space.has_random_parameter,
        TuneCompileError(
            "space can't contain random parameters, "
            "use sample method before calling this function"
        ),
    )
    dag = FugueWorkflow()
    dataset = TUNE_OBJECT_FACTORY.make_dataset(
        dag,
        space,
        df=train_df,
        partition_keys=partition_keys,
        temp_path=temp_path,
    )
    study = optimize_by_continuous_asha(
        objective=objective,
        dataset=dataset,
        plan=plan,
        checkpoint_path=temp_path,
        monitor=monitor,
    )
    study.result(top_n).yield_dataframe_as("result")

    return _run(
        dag=dag,
        execution_engine=execution_engine,
        execution_engine_conf=execution_engine_conf,
    )


def _run(
    dag: FugueWorkflow, execution_engine: Any, execution_engine_conf: Any
) -> List[TrialReport]:
    try:
        rows = list(
            dag.run(
                execution_engine,
                conf=execution_engine_conf,
            )["result"].as_dict_iterable()
        )
        return [
            TrialReport.from_jsondict(json.loads(r[TUNE_REPORT]))
            for r in sorted(rows, key=lambda r: r[TUNE_REPORT_METRIC])
        ]
    except FugueDataFrameError as e:
        raise e.__cause__ or e.__context__ or e



from typing import Any

from hyperopt.early_stop import no_progress_loss
from tune.noniterative.objective import NonIterativeObjectiveLocalOptimizer
from tune_test.local_optmizer import NonIterativeObjectiveLocalOptimizerTests

from tune_hyperopt import HyperoptLocalOptimizer


class HyperoptLocalOptimizerTests(NonIterativeObjectiveLocalOptimizerTests.Tests):
    def make_optimizer(self, **kwargs: Any) -> NonIterativeObjectiveLocalOptimizer:
        kwargs = {"seed": 0, "kwargs_func": _add_conf, **kwargs}
        return HyperoptLocalOptimizer(**kwargs)


def _add_conf(func, trial):
    return dict(early_stop_fn=no_progress_loss(50))



from sklearn.datasets import load_iris
from sklearn.linear_model import LogisticRegression
from tune import RandInt, Trial
from tune.constants import SPACE_MODEL_NAME, TUNE_DATASET_DF_DEFAULT_NAME
from tune.noniterative.objective import validate_noniterative_objective
from tune_hyperopt import HyperoptLocalOptimizer

from tune_sklearn.objective import SKCVObjective


def test_objective(tmpdir):
    dfs = load_iris(as_frame=True)
    df = dfs["data"]
    df["label"] = dfs["target"]
    df = df[df.label <= 1]

    t = Trial(
        "x",
        params={
            "max_iter": RandInt(2, 4),
            SPACE_MODEL_NAME: "sklearn.linear_model.LogisticRegression",
        },
        dfs={TUNE_DATASET_DF_DEFAULT_NAME: df},
    )
    obj = SKCVObjective(scoring="accuracy")
    optimizer = HyperoptLocalOptimizer(5, 0)

    def v(report):
        print(report.jsondict)
        assert report.sort_metric < 0
        assert "cv_scores" in report.metadata
        # assert report.trial.params["max_iter"] >= 2

    validate_noniterative_objective(obj, t, v, optimizer=optimizer)

    obj = SKCVObjective(scoring="accuracy", checkpoint_path=str(tmpdir))

    def v2(report):
        print(report.jsondict)
        assert report.sort_metric < 0
        assert "cv_scores" in report.metadata
        assert "checkpoint_path" in report.metadata
        # assert report.trial.params["max_iter"] >= 2

    validate_noniterative_objective(obj, t, v2, optimizer=optimizer)



import json
from typing import Any, List
from uuid import uuid4

from fs.base import FS as FSBase
from triad import assert_or_throw

_CHECKPOINT_STATE_FILE = "STATE"


class Checkpoint(object):
    def __init__(self, fs: FSBase):
        self._fs = fs
        try:
            self._iterations: List[str] = json.loads(
                fs.readtext(_CHECKPOINT_STATE_FILE)
            )
        except Exception:
            self._iterations = []

    def __len__(self) -> int:
        return len(self._iterations)

    @property
    def latest(self) -> FSBase:
        assert_or_throw(len(self) > 0, "checkpoint history is empty")
        return self._fs.opendir(self._iterations[-1])

    def create(self) -> "NewCheckpoint":
        return NewCheckpoint(self)


class NewCheckpoint(object):
    def __init__(self, checkpoint: Checkpoint):
        self._parent = checkpoint
        self._name = str(uuid4())

    def __enter__(self) -> FSBase:
        return self._parent._fs.makedir(self._name)

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        if exc_type is not None:
            try:
                self._parent._fs.removetree(self._name)
            except Exception:  # pragma: no cover
                pass
        else:
            new_iterations = self._parent._iterations + [self._name]
            self._parent._fs.writetext(
                _CHECKPOINT_STATE_FILE, json.dumps(new_iterations)
            )
            self._parent._iterations = new_iterations



from typing import Any, List, Optional, Union

import numpy as np

_IGNORABLE_ERROR = 1e-8


def uniform_to_continuous(
    value: Any, low: float, high: float, log: bool = False, base: Optional[float] = None
) -> Any:
    if low >= high:
        return low if np.isscalar(value) else np.full(np.shape(value))
    if not log:
        return value * (high - low) + low
    if base is None:
        ll, lh = np.log(low), np.log(high)
        return np.exp(value * (lh - ll) + ll)
    else:
        b = np.log(base)
        ll, lh = np.log(low) / b, np.log(high) / b
        return np.power(base, value * (lh - ll) + ll)


def uniform_to_discrete(
    value: Any,
    low: float,
    high: float,
    q: float,
    log: bool = False,
    include_high: bool = True,
    base: Optional[float] = None,
) -> Any:
    if low >= high:
        return low if np.isscalar(value) else np.full(np.shape(value))
    _high = adjust_high(low, high, q, include_high=include_high)
    _value = uniform_to_continuous(value, low, _high, log=log, base=base)
    return np.floor((_value - low) / q) * q + low


def uniform_to_integers(
    value: Any,
    low: int,
    high: int,
    q: int = 1,
    log: bool = False,
    include_high: bool = True,
    base: Optional[float] = None,
) -> Union[int, List[int]]:
    res = np.round(
        uniform_to_discrete(
            value, low, high, q=q, log=log, include_high=include_high, base=base
        )
    )
    if np.isscalar(res):
        return int(res)
    return [int(x) for x in res]


def uniform_to_choice(
    value: Any,
    choices: List[Any],
    log: bool = False,
    base: Optional[float] = None,
) -> Any:
    idx = uniform_to_integers(
        value, 1, len(choices), log=log, include_high=True, base=base
    )
    if isinstance(idx, int):
        return choices[idx - 1]
    return [choices[x - 1] for x in idx]


def normal_to_continuous(value: Any, mean: float, sigma: float) -> Any:
    return value * sigma + mean


def normal_to_discrete(value: Any, mean: float, sigma: float, q: float) -> Any:
    return np.round(value * sigma / q) * q + mean


def normal_to_integers(
    value: Any, mean: int, sigma: float, q: int = 1
) -> Union[int, List[int]]:
    res = normal_to_discrete(value, mean=mean, sigma=sigma, q=q)
    if np.isscalar(res):
        return int(res)
    return [int(x) for x in res]


def adjust_high(low: float, high: float, q: float, include_high: bool):
    _high = low + np.floor((high - low) / q + _IGNORABLE_ERROR) * q
    if abs(_high - high) < _IGNORABLE_ERROR:
        if include_high:
            _high = high + q
    else:
        _high += q
    return _high



from typing import Any, Type, Dict

from triad.utils.convert import get_full_type_path, to_type

from tune_tensorflow.spec import KerasTrainingSpec
from tune import Space
from tune.constants import SPACE_MODEL_NAME

_TYPE_DICT: Dict[str, Type[KerasTrainingSpec]] = {}


def to_keras_spec(obj: Any) -> Type[KerasTrainingSpec]:
    if isinstance(obj, str) and obj in _TYPE_DICT:
        return _TYPE_DICT[obj]
    return to_type(obj, KerasTrainingSpec)


def to_keras_spec_expr(spec: Any) -> str:
    if isinstance(spec, str):
        spec = to_keras_spec(spec)
    return get_full_type_path(spec)


def keras_space(model: Any, **params: Any) -> Space:
    expr = to_keras_spec_expr(model)
    _TYPE_DICT[expr] = to_keras_spec(model)
    data = {SPACE_MODEL_NAME: expr}
    data.update(params)
    return Space(**data)


def extract_keras_spec(
    params: Dict[str, Any], type_dict: Dict[str, Any]
) -> Type[KerasTrainingSpec]:
    obj = params[SPACE_MODEL_NAME]
    if isinstance(obj, str) and obj in type_dict:
        return type_dict[obj]
    return to_keras_spec(obj)



import json

import numpy as np
from pytest import raises
from scipy import stats
from triad import to_uuid
from tune.concepts.space import (
    Choice,
    Grid,
    NormalRand,
    NormalRandInt,
    Rand,
    RandInt,
    Space,
)
from tune.concepts.space.parameters import decode_params
from tune._utils import assert_close


def test_grid():
    v = Grid("a", "b")
    assert ["a", "b"] == list(v)

    v2 = Grid("b", "a")
    assert v == v and v != v2
    assert to_uuid(v) != to_uuid(v2)

    raises(ValueError, lambda: Grid())


def test_choice():
    raises(ValueError, lambda: Choice())
    v = Choice("a", "b", "c")
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    assert v.generate_many(20, 0) == v.generate_many(20, 0)
    assert v.generate_many(20, 0) != v.generate_many(20, 1)
    actual = set(v.generate_many(20, 0))
    assert set(["a", "b", "c"]) == actual

    assert to_uuid(v) != to_uuid(Grid("a", "b", "c"))
    assert v != Grid("a", "b", "c")

    v = Choice(1, 2, 3)
    assert json.loads(json.dumps({"x": v.generate(0)}))["x"] <= 3

    v = Choice("a", "b", "c")
    assert isinstance(json.loads(json.dumps({"x": v.generate(0)}))["x"], str)


def test_rand():
    with raises(ValueError):
        Rand(1.0, 0.9)

    with raises(ValueError):
        Rand(1.0, 10, q=-0.1)

    with raises(ValueError):
        Rand(1.0, 1.0, include_high=False)

    with raises(ValueError):
        Rand(0.0, 1.0, log=True)  # for log, low>=1.0

    v = Rand(1.0, 1.0, q=0.1, log=False)
    assert 1.0 == v.generate()
    assert 1.0 == v.generate(10)

    v = Rand(1.0, 1.0, q=0.1, log=True)
    assert 1.0 == v.generate()
    assert 1.0 == v.generate(10)

    v = Rand(1.0, 2.0, q=0.1, log=False)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    assert_close([x / 10 for x in range(10, 21)], v.generate_many(100, 0))

    v = Rand(1.0, 2.09999, q=0.1, log=False)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    assert_close([x / 10 for x in range(10, 21)], v.generate_many(100, 0))

    v = Rand(1.0, 2.0, q=0.1, log=False, include_high=False)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    assert_close([x / 10 for x in range(10, 20)], v.generate_many(100, 0))

    v = Rand(1.0, 2.09999, q=0.1, log=False, include_high=False)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    assert_close([x / 10 for x in range(10, 21)], v.generate_many(100, 0))

    v = Rand(1.0, 2.0, log=True, include_high=False)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    res = v.generate_many(10000, 0)
    for x in res:
        assert x >= 1.0 and x <= 2.0
    t = stats.kstest(np.log(res), "uniform", args=(np.log(1), np.log(2) - np.log(1)))
    assert t.pvalue > 0.4


def test_randint():
    with raises(ValueError):
        RandInt(0, 10, log=True)  # for log, low>=1.0

    v = RandInt(10, 20, log=False)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    actual = set(v.generate_many(100, 0))
    assert set(range(10, 21)) == actual

    v = RandInt(10, 20, include_high=False)
    actual = set(v.generate_many(100, 0))
    assert set(range(10, 20)) == actual

    v = RandInt(10, 20, q=5, include_high=False)
    actual = set(v.generate_many(100, 0))
    assert set([10, 15]) == actual

    v = RandInt(10, 20, q=5, include_high=True)
    actual = set(v.generate_many(100, 0))
    assert set([10, 15, 20]) == actual

    v = RandInt(3, 20, log=True)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    actual = set(v.generate_many(1000, 0))
    assert set(range(3, 21)) == actual


def test_normal_rand():
    with raises(ValueError):
        NormalRand(1.0, 0.0)

    with raises(ValueError):
        NormalRand(1.0, -1.0)

    v = NormalRand(0.05, 0.2)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    res = v.generate_many(100000, 0)
    t = stats.kstest(res, "norm", args=(0.05, 0.2))
    assert t.pvalue > 0.4

    v = NormalRand(0.05, 0.2, q=0.1)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    actual = [x for x in v.generate_many(1000, 0) if x >= -0.155 and x <= 0.255]
    assert_close([-0.15, -0.05, 0.05, 0.15, 0.25], actual)


def test_normal_randint():
    v = NormalRandInt(5, 2)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(1)
    actual = set(v.generate_many(50, 0))
    for x in [3, 4, 5, 6, 7]:
        assert x in actual

    v = NormalRandInt(5, 2, q=3)
    assert v.generate(0) == v.generate(0)
    assert v.generate(0) != v.generate(2)
    actual = set(v.generate_many(50, 0))
    for x in [-1, 2, 5, 8, 11]:
        assert x in actual
    assert 6 not in actual


def test_encode_decode_params():
    s1 = Space(
        a=Grid(1, 2),
        b=Rand(1.0, 7.1, 0.2, log=True),
        c=Choice(1, 2, 3),
        d=[Grid(1, 2), Rand(0, 2.0)],
        e={"x": "xx", "y": Choice("a", "b")},
        f=RandInt(0, 10, log=False),
        g=NormalRand(0.1, 1.0, q=0.1),
        h=NormalRandInt(0.1, 1.0),
    )
    actual = [decode_params(x) for x in s1.encode()]
    assert list(s1) == actual
    for x in s1.encode():
        print(json.dumps(x, indent=2))



import os

from setuptools import find_packages, setup

from tune_version import __version__

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()


def get_version() -> str:
    tag = os.environ.get("RELEASE_TAG", "")
    if "dev" in tag.split(".")[-1]:
        return tag
    if tag != "":
        assert tag == __version__, "release tag and version mismatch"
    return __version__


setup(
    name="tune",
    version=get_version(),
    packages=find_packages(),
    description="An abstraction layer for hyper parameter tuning",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    author="Han Wang",
    author_email="goodwanghan@gmail.com",
    keywords="hyper parameter hyperparameter tuning tune tuner",
    url="http://github.com/fugue-project/tune",
    install_requires=["fugue>=0.5.6", "cloudpickle"],
    extras_require={
        "hyperopt": ["hyperopt"],
        "optuna": ["optuna"],
        "visual": ["seaborn", "plotly"],
        "tensorflow": ["tensorflow"],
        "notebook": ["fugue[notebook]", "seaborn", "plotly"],
        "sklearn": ["scikit-learn"],
        "all": [
            "hyperopt",
            "optuna",
            "seaborn",
            "tensorflow",
            "fugue[notebook]",
            "plotly",
            "scikit-learn",
        ],
    },
    classifiers=[
        # "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires=">=3.6",
)



import json

from pytest import raises

from tune import Choice, Grid, Rand, RandInt, Space
from tune.concepts.space import HorizontalSpace, VerticalSpace


def test_single_space():
    space = Space(a=1, b=Grid(2, 3, 4))
    dicts = list(space)
    dicts = list(space)  # make sure iteration doesn't alter the object
    assert 3 == len(dicts)
    assert dict(a=1, b=2) == dicts[0]
    assert dict(a=1, b=3) == dicts[1]

    dicts = list(Space(a=Grid(None, "x"), b=Grid(2, 3)))
    assert 4 == len(dicts)

    dicts = list(Space(a=1, b=[Grid(2, 3), Grid(4, 5)]))
    assert 4 == len(dicts)
    assert dict(a=1, b=[2, 4]) == dicts[0]
    assert dict(a=1, b=[2, 5]) == dicts[1]
    assert dict(a=1, b=[3, 4]) == dicts[2]
    assert dict(a=1, b=[3, 5]) == dicts[3]

    dicts = list(Space(a=1, b=dict(x=Grid(2, 3), y=Grid(4, 5))))
    assert 4 == len(dicts)
    assert dict(a=1, b=dict(x=2, y=4)) == dicts[0]
    assert dict(a=1, b=dict(x=2, y=5)) == dicts[1]
    assert dict(a=1, b=dict(x=3, y=4)) == dicts[2]
    assert dict(a=1, b=dict(x=3, y=5)) == dicts[3]


def test_single_space_sample():
    assert not Space(a=1).has_random_parameter
    assert not Space(a=1, b=Grid(1, 2)).has_random_parameter
    assert Space(a=1, b=[Grid(1, 2), Rand(0.0, 1.0)]).has_random_parameter

    dicts = list(Space(a=1, b=Grid(1, 2)).sample(100))
    assert 2 == len(dicts)

    space = Space(a=1, b=[Grid(1, 2), Rand(0.0, 1.0)], c=Choice("a", "b"))
    ec = list(space.encode())
    assert list(space.sample(5, 0)) == list(space.sample(5, 0))
    assert list(space.sample(5, 0)) != list(space.sample(5, 1))
    dicts = list(space.sample(5, 0))
    assert 10 == len(dicts)
    assert 5 == len(set(d["b"][1] for d in dicts))
    assert ec == list(space.encode())  # make sure the instance value was not altered


def test_horizontal_space():
    spaces = list(HorizontalSpace())
    assert 1 == len(spaces)
    assert {} == spaces[0]

    spaces = list(HorizontalSpace(a=10, b=[1, 2], c=dict(x=1)))
    assert 1 == len(spaces)
    assert dict(a=10, b=[1, 2], c=dict(x=1)) == spaces[0]

    spaces = list(HorizontalSpace(dict(a=10, b=[1, 2], c=dict(x=1))))
    assert 1 == len(spaces)
    assert dict(a=10, b=[1, 2], c=dict(x=1)) == spaces[0]

    spaces = list(HorizontalSpace(dict(a=10), dict(b=[1, 2], c=dict(x=1))))
    assert 1 == len(spaces)
    assert dict(a=10, b=[1, 2], c=dict(x=1)) == spaces[0]

    raises(ValueError, lambda: HorizontalSpace(10))

    assert not HorizontalSpace(dict(a=10, b=[1, 2], c=dict(x=1))).has_random_parameter

    space = HorizontalSpace(
        Space(a=Grid(1, 2), b=Rand(0.0, 1.0)), Space(c=Choice(3, 4))
    )
    ec = list(space.encode())
    assert space.has_random_parameter
    assert list(space.sample(5, 0)) == list(space.sample(5, 0))
    assert list(space.sample(5, 0)) != list(space.sample(5, 1))
    dicts = list(space.sample(5, 0))
    assert 10 == len(dicts)
    assert 5 == len(set(d["b"] for d in dicts))
    assert ec == list(space.encode())


def test_vertical_space():
    spaces = list(VerticalSpace())
    assert 0 == len(spaces)

    spaces = list(VerticalSpace(dict(a=10)))
    assert [dict(a=10)] == spaces

    spaces = list(VerticalSpace(dict(a=10), [dict(b=11), dict(c=12)]))
    assert [dict(a=10), dict(b=11), dict(c=12)] == spaces

    spaces = list(VerticalSpace(HorizontalSpace(a=10), dict(b=10)))
    assert [dict(a=10), dict(b=10)] == spaces

    raises(ValueError, lambda: VerticalSpace(10))

    assert not VerticalSpace(HorizontalSpace(a=10), dict(b=10)).has_random_parameter

    space = VerticalSpace(Space(a=Grid(1, 2), b=Rand(0.0, 1.0)), Space(c=Choice(3, 4)))
    ec = list(space.encode())
    assert space.has_random_parameter
    assert list(space.sample(5, 0)) == list(space.sample(5, 0))
    assert list(space.sample(5, 0)) != list(space.sample(5, 1))
    dicts = list(space.sample(5, 0))
    assert 15 == len(dicts)
    assert 5 == len(set(d["b"] for d in dicts if "b" in d))
    assert ec == list(space.encode())


def test_space_combo():
    spaces = list(HorizontalSpace(dict(a=10), []))
    assert [dict(a=10)] == spaces

    spaces = list(HorizontalSpace(dict(a=10), [dict(b=20), dict(c=30, a=11)]))
    assert 2 == len(spaces)
    assert dict(a=10, b=20) == spaces[0]
    assert dict(a=11, c=30) == spaces[1]

    spaces = list(
        HorizontalSpace(
            HorizontalSpace(a=10),
            VerticalSpace(dict(b=20), HorizontalSpace(c=30, a=None)),
        )
    )
    assert 2 == len(spaces)
    assert dict(a=10, b=20) == spaces[0]
    assert dict(a=None, c=30) == spaces[1]

    spaces = list(
        HorizontalSpace(
            dict(a=HorizontalSpace(dict(aa=10), VerticalSpace(dict(), dict(cc=12)))),
            VerticalSpace(dict(b=20), HorizontalSpace(c=30)),
        )
    )
    assert 4 == len(spaces)
    assert dict(a=dict(aa=10), b=20) == spaces[0]
    assert dict(a=dict(aa=10, cc=12), b=20) == spaces[1]
    assert dict(a=dict(aa=10), c=30) == spaces[2]
    assert dict(a=dict(aa=10, cc=12), c=30) == spaces[3]

    spaces = list(
        HorizontalSpace(a=VerticalSpace(HorizontalSpace(x=[1, 2]), dict(y=None)))
    )
    assert 2 == len(spaces)
    assert dict(a=dict(x=[1, 2])) == spaces[0]
    assert dict(a=dict(y=None)) == spaces[1]


def test_operators():
    s1 = Space(a=1, b=Grid(2, 3))
    s2 = Space(c=Grid("a", "b"))
    assert [
        dict(a=1, b=2, c="a"),
        dict(a=1, b=2, c="b"),
        dict(a=1, b=3, c="a"),
        dict(a=1, b=3, c="b"),
    ] == list(s1 * s2)

    assert [
        dict(a=1, b=2),
        dict(a=1, b=3),
        dict(c="a"),
        dict(c="b"),
    ] == list(s1 + s2)

    assert [
        dict(a=1, b=2, c="a"),
        dict(a=1, b=3, c="a"),
        dict(a=1, b=2, c="b"),
        dict(a=1, b=3, c="b"),
    ] == list(s1 * [dict(c="a"), dict(c="b")])

    assert [
        dict(a=1, b=2),
        dict(a=1, b=3),
        dict(c="a"),
        dict(c="b"),
    ] == list(s1 + [dict(c="a"), dict(c="b")])

    s1 = Space(a=1, b=Grid(2, 3))
    s2 = Space(c=Grid("a", "b"))
    s3 = Space(d=5)
    assert (
        [
            dict(a=1, b=2),
            dict(a=1, b=3),
            dict(c="a"),
            dict(c="b"),
            dict(d=5),
        ]
        == list(sum([s1, s2, s3]))
        == list(sum([s1, s2, s3], None))
    )



from tune import Trial, TrialReport

from tune_notebook import (
    NotebookSimpleChart,
    NotebookSimpleHist,
    NotebookSimpleRungs,
    NotebookSimpleTimeSeries,
    PrintBest,
)
from tune_notebook.monitors import _ReportBin


def test_report_bin():
    t1 = Trial("a", dict(a=1, b=2), keys=["x", "y"])
    r1 = TrialReport(t1, 0.8, sort_metric=-0.8)
    t2 = Trial("b", dict(a=11, b=12), keys=["xx", "y"])
    r2 = TrialReport(t2, 0.7, sort_metric=-0.7)
    t3 = Trial("c", dict(a=10, b=20), keys=["x", "y"])
    r3 = TrialReport(t3, 0.9, sort_metric=-0.9)

    b = _ReportBin()
    assert 0 == len(b.records)
    assert b.best is None
    assert b.on_report(r1)
    assert b.on_report(r3)
    assert r3 is b.best
    assert 2 == len(b.records)

    b = _ReportBin()
    assert b.on_report(r3)
    assert not b.on_report(r1)
    assert r3 is b.best
    assert 2 == len(b.records)

    b = _ReportBin(new_best_only=True)
    assert b.on_report(r3)
    assert not b.on_report(r1)
    assert r3 is b.best
    assert 1 == len(b.records)


def test_print_best():
    t1 = Trial("a", dict(a=1, b=2), keys=["x", "y"])
    r1 = TrialReport(t1, 0.8, sort_metric=-0.8)
    t2 = Trial("b", dict(a=11, b=12), keys=["xx", "y"])
    r2 = TrialReport(t2, 0.7, sort_metric=-0.7)
    t3 = Trial("c", dict(a=10, b=20), keys=["x", "y"])
    r3 = TrialReport(t3, 0.9, sort_metric=-0.9)

    b = PrintBest()
    b.on_report(r3)
    b.on_report(r2)
    b.on_report(r1)


def _test_charts():
    t1 = Trial("a", dict(a=1, b=2), keys=["x", "y"])
    r1 = TrialReport(t1, 0.8, sort_metric=-0.8)
    t2 = Trial("b", dict(a=11, b=12), keys=["xx", "y"])
    r2 = TrialReport(t2, 0.7, sort_metric=-0.7)
    t3 = Trial("c", dict(a=10, b=20), keys=["x", "y"])
    r3 = TrialReport(t3, 0.9, sort_metric=-0.9)

    b = NotebookSimpleTimeSeries("1sec")
    b.on_report(r3)
    b.on_report(r2)
    b.on_report(r1)

    b = NotebookSimpleHist("2sec")
    b.on_report(r3)
    b.on_report(r2)
    b.on_report(r1)

    b = NotebookSimpleRungs("3sec")
    b.on_report(r3)
    b.on_report(r2)
    b.on_report(r1)



import os
from typing import Any, Callable, List, Optional, Tuple
from uuid import uuid4

from triad import FileSystem
from tune.api.factory import TUNE_OBJECT_FACTORY
from tune.concepts.dataset import StudyResult, TuneDataset
from tune.concepts.flow import TrialReport
from tune.iterative.asha import ASHAJudge, RungHeap
from tune.iterative.sha import _NonIterativeObjectiveWrapper
from tune.iterative.study import IterativeStudy
from tune.noniterative.study import NonIterativeStudy


def optimize_noniterative(
    objective: Any,
    dataset: TuneDataset,
    optimizer: Any = None,
    distributed: Optional[bool] = None,
    monitor: Any = None,
    stopper: Any = None,
    stop_check_interval: Any = None,
) -> StudyResult:
    _objective = TUNE_OBJECT_FACTORY.make_noniterative_objective(objective)
    _optimizer = TUNE_OBJECT_FACTORY.make_noniterative_local_optimizer(optimizer)
    _stopper = TUNE_OBJECT_FACTORY.make_noniterative_stopper(stopper)
    _monitor = TUNE_OBJECT_FACTORY.make_monitor(monitor)
    study = NonIterativeStudy(_objective, _optimizer)
    return study.optimize(
        dataset,
        distributed=distributed,
        monitor=_monitor,
        stopper=_stopper,
        stop_check_interval=stop_check_interval,
    )


def optimize_by_sha(
    objective: Any,
    dataset: TuneDataset,
    plan: List[Tuple[float, int]],
    checkpoint_path: str = "",
    distributed: Optional[bool] = None,
    monitor: Any = None,
) -> StudyResult:
    _objective = TUNE_OBJECT_FACTORY.make_iterative_objective(objective)
    _monitor = TUNE_OBJECT_FACTORY.make_monitor(monitor)
    checkpoint_path = TUNE_OBJECT_FACTORY.get_path_or_temp(checkpoint_path)
    path = os.path.join(checkpoint_path, str(uuid4()))
    for budget, keep in plan:
        obj = _NonIterativeObjectiveWrapper(
            _objective, checkpoint_path=path, budget=budget
        )
        result = optimize_noniterative(
            obj, dataset, distributed=distributed, monitor=_monitor
        )
        dataset = result.next_tune_dataset(keep)
    return result


def optimize_by_hyperband(
    objective: Any,
    dataset: TuneDataset,
    plans: List[List[Tuple[float, int]]],
    checkpoint_path: str = "",
    distributed: Optional[bool] = None,
    monitor: Any = None,
) -> StudyResult:
    _monitor = TUNE_OBJECT_FACTORY.make_monitor(monitor)
    weights = [float(p[0][1]) for p in plans]
    datasets = dataset.divide(weights, seed=0)
    result: Any = None
    for d, plan in zip(datasets, plans):
        r = optimize_by_sha(
            objective=objective,
            dataset=d,
            plan=plan,
            checkpoint_path=checkpoint_path,
            distributed=distributed,
            monitor=_monitor,
        )
        if result is None:
            result = r
        else:
            result.union_with(r)
    return result


def optimize_by_continuous_asha(
    objective: Any,
    dataset: TuneDataset,
    plan: List[Tuple[float, int]],
    checkpoint_path: str = "",
    always_checkpoint: bool = False,
    study_early_stop: Optional[Callable[[List[Any], List[RungHeap]], bool]] = None,
    trial_early_stop: Optional[
        Callable[[TrialReport, List[TrialReport], List[RungHeap]], bool]
    ] = None,
    monitor: Any = None,
) -> StudyResult:
    _objective = TUNE_OBJECT_FACTORY.make_iterative_objective(objective)
    _monitor = TUNE_OBJECT_FACTORY.make_monitor(monitor)
    checkpoint_path = TUNE_OBJECT_FACTORY.get_path_or_temp(checkpoint_path)
    judge = ASHAJudge(
        schedule=plan,
        always_checkpoint=always_checkpoint,
        study_early_stop=study_early_stop,
        trial_early_stop=trial_early_stop,
        monitor=_monitor,
    )
    path = os.path.join(checkpoint_path, str(uuid4()))
    FileSystem().makedirs(path, recreate=True)
    study = IterativeStudy(_objective, checkpoint_path=path)
    return study.optimize(dataset, judge=judge)



from datetime import datetime
from time import sleep

from pytest import raises
from tune.concepts.flow.report import TrialReport
from tune.concepts.flow.trial import Trial
from tune.noniterative.stopper import (
    SimpleNonIterativeStopper,
    n_samples,
    n_updates,
    no_update_period,
    small_improvement,
)


def test_simple_stopper():
    r1 = mr([], 0.2)
    r2 = mr(["x"], 0.4)
    r3 = mr([], 0.1)
    r4 = mr(["x"], 0.5)

    s = MockSimpleStopper(lambda latest, updated, reports: len(reports) >= 1)
    assert 0 == len(s.get_reports(r1.trial))
    assert s.can_accept(r1.trial)
    s.judge(r1)
    assert s._last is r1

    assert s.can_accept(r2.trial)
    s.judge(r2)
    assert s._last is r2

    assert not s.can_accept(r3.trial)
    assert not s.can_accept(r4.trial)


def test_logical_ops():
    r1 = mr([], 0.5)
    r2 = mr([], 0.4)
    r3 = mr([], 0.3)
    r4 = mr([], 0.2)
    take_two = MockSimpleStopper(lambda latest, updated, reports: len(reports) >= 2)
    ends_small = MockSimpleStopper(
        lambda latest, updated, reports: reports[-1].sort_metric <= 0.3
    )

    s = take_two & ends_small
    assert s.can_accept(r1.trial)
    s.judge(r1)
    assert s.can_accept(r2.trial)
    s.judge(r2)
    assert s.can_accept(r3.trial)
    s.judge(r3)
    assert not s.can_accept(r4.trial)

    with raises(AssertionError):
        take_two | ends_small  # can't reuse updated stoppers

    take_two = MockSimpleStopper(lambda latest, updated, reports: len(reports) >= 2)
    ends_small = MockSimpleStopper(
        lambda latest, updated, reports: reports[-1].sort_metric <= 0.3
    )

    s = take_two | ends_small
    assert s.can_accept(r1.trial)
    s.judge(r1)
    assert s.can_accept(r2.trial)
    s.judge(r2)
    assert not s.can_accept(r3.trial)


def test_n_samples():
    r1 = mr([], 0.1)
    r2 = mr([], 0.4)
    r3 = mr([], 0.3)

    s = n_samples(2)
    assert s.can_accept(r1.trial)
    s.judge(r1)
    assert s.can_accept(r2.trial)
    s.judge(r2)
    assert not s.can_accept(r3.trial)


def test_n_updates():
    r1 = mr([], 0.4)
    r2 = mr([], 0.5)
    r3 = mr([], 0.3)
    r4 = mr([], 0.2)

    s = n_updates(2)
    assert s.can_accept(r1.trial)
    s.judge(r1)
    assert s.can_accept(r2.trial)
    s.judge(r2)
    assert s.can_accept(r3.trial)
    s.judge(r3)  # updates satisfied
    assert not s.can_accept(r4.trial)


def test_no_update_period():
    r1 = mr([], 0.1)
    r2 = mr([], 0.4)
    r3 = mr([], 0.3)
    r4 = mr([], 0.3)

    s = no_update_period("0.2sec")
    assert s.can_accept(r1.trial)
    s.judge(r1)
    sleep(0.5)
    assert s.can_accept(r2.trial)
    s.judge(r2)  # not an update, and exceeded timeout, so trigger the shut down
    assert not s.can_accept(r3.trial)

    s = no_update_period("0.2sec")
    assert s.can_accept(r2.trial)
    s.judge(r2)
    sleep(0.5)
    assert s.can_accept(r3.trial)
    s.judge(r3)  # an update will not trigger shut down
    assert s.can_accept(r1.trial)
    s.judge(r1)
    assert s.can_accept(r4.trial)
    s.judge(r4)


def test_small_improvement():
    r1 = mr([], 0.5)
    r2 = mr([], 0.4)
    r22 = mr([], 0.51)
    r3 = mr([], 0.39)
    r4 = mr([], 0.2)

    s = small_improvement(0.09, 1)
    assert s.can_accept(r1.trial)
    s.judge(r1)
    assert s.can_accept(r2.trial)
    s.judge(r2)
    assert s.can_accept(r3.trial)
    s.judge(r3)
    assert not s.can_accept(r4.trial)

    s = small_improvement(0.25, 2)
    assert s.can_accept(r1.trial)
    s.judge(r1)
    assert s.can_accept(r2.trial)
    s.judge(r2)
    assert s.can_accept(r22.trial)
    s.judge(r22)  # if not an update, it doesn't count
    assert s.can_accept(r3.trial)
    s.judge(r3)
    assert not s.can_accept(r4.trial)


class MockSimpleStopper(SimpleNonIterativeStopper):
    def __init__(self, func):
        super().__init__(
            partition_should_stop=self.partition_should_stop, log_best_only=False
        )
        self._last = None
        self._func = func

    def partition_should_stop(self, latest_report, updated, reports) -> bool:
        self._last = latest_report
        return self._func(latest_report, updated, reports)


def mr(keys, metric):
    t = Trial(str(metric), dict(a=1), keys=keys)
    return TrialReport(t, metric, log_time=datetime.now())



from time import sleep
from typing import List

import pandas as pd
from fugue import FugueWorkflow
from pytest import raises
from tune import optimize_noniterative, suggest_for_noniterative_objective
from tune.concepts.dataset import TuneDatasetBuilder
from tune.concepts.flow import Monitor
from tune.concepts.space import Grid, Space
from tune.constants import TUNE_REPORT, TUNE_REPORT_METRIC
from tune.exceptions import TuneInterrupted
from tune.noniterative.convert import to_noniterative_objective
from tune.noniterative.stopper import n_samples


def objective(a: float, b: pd.DataFrame) -> float:
    return a ** 2 + b.shape[0]


def objective2(a: float, b: pd.DataFrame) -> float:
    return -(a ** 2 + b.shape[0])


def objective3(a: float, b: pd.DataFrame) -> float:
    if a == -2:
        raise TuneInterrupted()
    return a ** 2 + b.shape[0]


def objective4(a: float, b: pd.DataFrame) -> float:
    if a == -2:
        raise ValueError("expected")
    return a ** 2 + b.shape[0]


def assert_metric(df: pd.DataFrame, metrics: List[float]) -> None:
    assert len(metrics) == df.shape[0]
    assert set(metrics) == set(df[TUNE_REPORT_METRIC].tolist())


class M(Monitor):
    def __init__(self):
        super().__init__()
        self._reports = []

    def on_report(self, report) -> None:
        self._reports.append(report)


def test_study(tmpdir):
    space = Space(a=Grid(-2, 0, 1))
    input_df = pd.DataFrame([[0, 1], [1, 1], [0, 2]], columns=["a", "b"])
    dag = FugueWorkflow()
    monitor = M()

    # no data partition
    builder = TuneDatasetBuilder(space, str(tmpdir)).add_df("b", dag.df(input_df))
    dataset = builder.build(dag, 1)
    for distributed in [True, False, None]:
        # min_better = True
        result = optimize_noniterative(
            objective=objective,
            dataset=dataset,
            distributed=distributed,
        )
        result.result()[[TUNE_REPORT, TUNE_REPORT_METRIC]].output(
            assert_metric, params=dict(metrics=[3.0, 4.0, 7.0])
        )
        result.result(2)[[TUNE_REPORT, TUNE_REPORT_METRIC]].output(
            assert_metric, params=dict(metrics=[3.0, 4.0])
        )

        # min_better = False
        result = optimize_noniterative(
            objective=to_noniterative_objective(objective, min_better=False),
            dataset=dataset,
            distributed=distributed,
        )
        result.result()[[TUNE_REPORT, TUNE_REPORT_METRIC]].output(
            assert_metric, params=dict(metrics=[-7.0, -4.0, -3.0])
        )
        result.result(2)[[TUNE_REPORT, TUNE_REPORT_METRIC]].output(
            assert_metric, params=dict(metrics=[-7.0, -4.0])
        )

    # with data partition
    builder = TuneDatasetBuilder(space, str(tmpdir)).add_df(
        "b", dag.df(input_df).partition_by("a")
    )
    dataset = builder.build(dag, 1)
    for distributed in [True, False, None]:
        result = optimize_noniterative(
            objective=to_noniterative_objective(objective),
            dataset=dataset,
            distributed=distributed,
            monitor=monitor,
        )
        result.result()[[TUNE_REPORT, TUNE_REPORT_METRIC]].output(
            assert_metric, params=dict(metrics=[2.0, 3.0, 6.0, 1.0, 2.0, 5.0])
        )
        result.result(1)[[TUNE_REPORT, TUNE_REPORT_METRIC]].output(
            assert_metric, params=dict(metrics=[1.0, 2.0])
        )

    dag.run()

    assert 3 * 3 * 2 == len(monitor._reports)


def test_study_with_stopper(tmpdir):
    space = Space(a=Grid(-2, 0, 1))
    input_df = pd.DataFrame([[0, 1], [1, 1], [0, 2]], columns=["a", "b"])

    result = suggest_for_noniterative_objective(
        objective=objective,
        space=space,
        df=input_df,
        df_name="b",
        stopper=n_samples(2),
        top_n=0,
        shuffle_candidates=False,
        temp_path=str(tmpdir),
    )
    assert [3.0, 7.0] == [x.metric for x in result]

    monitor = M()
    result = suggest_for_noniterative_objective(
        objective=objective,
        space=space,
        df=input_df,
        df_name="b",
        stopper=n_samples(2),
        monitor=monitor,
        top_n=0,
        shuffle_candidates=False,
        temp_path=str(tmpdir),
    )
    assert [3.0, 7.0] == [x.metric for x in result]

    assert 2 == len(monitor._reports)

    # Tune interrupted
    monitor = M()
    result = suggest_for_noniterative_objective(
        objective=objective3,
        space=space,
        df=input_df,
        df_name="b",
        stopper=n_samples(2),
        monitor=monitor,
        top_n=0,
        shuffle_candidates=False,
        temp_path=str(tmpdir),
    )
    assert [3.0, 4.0] == [x.metric for x in result]

    assert 2 == len(monitor._reports)


def test_study_with_exceptions(tmpdir):
    space = Space(a=Grid(-2, 0, 1))
    input_df = pd.DataFrame([[0, 1], [1, 1], [0, 2]], columns=["a", "b"])

    with raises(ValueError):
        suggest_for_noniterative_objective(
            objective=objective4,
            space=space,
            df=input_df,
            df_name="b",
            temp_path=str(tmpdir),
        )

    with raises(ValueError):
        monitor = M()
        suggest_for_noniterative_objective(
            objective=objective4,
            space=space,
            df=input_df,
            df_name="b",
            stopper=n_samples(2),
            monitor=monitor,
            temp_path=str(tmpdir),
        )



from time import sleep

from pytest import raises
from triad.utils.convert import to_timedelta
from tune._utils import run_monitored_process
from tune.exceptions import TuneInterrupted


def test_run_monitored_process():
    # happy case
    assert 10 == run_monitored_process(t1, [1], {}, lambda: True, "5sec")
    # stop before responding
    with raises(TuneInterrupted):
        run_monitored_process(t1, [1], dict(wait="20sec"), lambda: True, "0.2sec")
    # non stop before responding
    assert 10 == run_monitored_process(
        t1, [1], dict(wait="1sec"), lambda: False, "0.2sec"
    )
    with raises(NotImplementedError):
        run_monitored_process(t2, [], {}, lambda: True, "5sec")
    assert run_monitored_process(t3, [], {}, lambda: True, "5sec") is None


def t1(a, wait=None):
    if wait is not None:
        sleep(to_timedelta(wait).total_seconds())
    return a * 10


def t2():
    raise NotImplementedError


def t3():
    pass



from datetime import datetime
from threading import RLock
from typing import Any, Dict, List

import pandas as pd
from triad.utils.convert import to_timedelta
from tune import Monitor, TrialReport, TrialReportLogger


class PrintBest(Monitor):
    def __init__(self):
        super().__init__()
        self._lock = RLock()
        self._bins: Dict[str, "_ReportBin"] = {}

    def on_report(self, report: TrialReport) -> None:
        with self._lock:
            key = str(report.trial.keys)
            if key not in self._bins:
                self._bins[key] = _ReportBin(new_best_only=True)
            rbin = self._bins[key]
        if rbin.on_report(report):
            print(report.trial.keys, report.metric, report.jsondict)


class NotebookSimpleChart(Monitor):
    def __init__(
        self,
        interval: Any = "1sec",
        best_only: bool = True,
        always_update: bool = False,
    ):
        super().__init__()
        self._lock = RLock()
        self._last: Any = None
        self._bins: Dict[str, "_ReportBin"] = {}
        self._interval = to_timedelta(interval)
        self._best_only = best_only
        self._always_update = always_update

    def on_report(self, report: TrialReport) -> None:
        now = datetime.now()
        with self._lock:
            key = str(report.trial.keys)
            if key not in self._bins:
                self._bins[key] = _ReportBin(new_best_only=self._best_only)
            rbin = self._bins[key]

        updated = rbin.on_report(report)
        if not updated and not self._always_update:
            return

        with self._lock:
            if self._last is None or now - self._last > self._interval:
                self._redraw()
                self._last = datetime.now()

    def plot(self, df: pd.DataFrame) -> None:
        return  # pragma: no cover

    def finalize(self) -> None:
        self._redraw()

    def _redraw(self) -> None:
        import matplotlib.pyplot as plt
        from IPython.display import clear_output

        df = pd.concat(
            [
                pd.DataFrame(
                    x.records,
                    columns=[
                        "partition",
                        "rung",
                        "time",
                        "id",
                        "metric",
                        "best_metric",
                    ],
                )
                for x in self._bins.values()
            ]
        )
        clear_output()
        self.plot(df)
        plt.show()
        for best in [x.best for x in self._bins.values() if x.best is not None]:
            if best is not None:
                print(best.trial.keys, best.metric, best.jsondict)


class NotebookSimpleRungs(NotebookSimpleChart):
    def __init__(self, interval: Any = "1sec"):
        super().__init__(interval, best_only=False, always_update=True)

    def plot(self, df: pd.DataFrame) -> None:
        import seaborn as sns

        sns.lineplot(data=df, x="rung", y="metric", hue="id", marker="o", legend=False)


class NotebookSimpleTimeSeries(NotebookSimpleChart):
    def __init__(self, interval: Any = "1sec"):
        super().__init__(interval)

    def plot(self, df: pd.DataFrame) -> None:
        import seaborn as sns

        sns.lineplot(data=df, x="time", y="best_metric", hue="partition", marker="o")


class NotebookSimpleHist(NotebookSimpleChart):
    def __init__(self, interval: Any = "1sec"):
        super().__init__(interval, best_only=False, always_update=True)

    def plot(self, df: pd.DataFrame) -> None:
        import seaborn as sns

        sns.histplot(data=df, x="metric", hue="partition")


class _ReportBin(TrialReportLogger):
    def __init__(self, new_best_only: bool = False):
        super().__init__(new_best_only=new_best_only)
        self._values: List[List[Any]] = []

    def log(self, report: TrialReport) -> None:
        self._values.append(
            [
                str(report.trial.keys),
                report.rung,
                datetime.now(),
                report.trial_id,
                report.metric,
                self.best.metric,  # type: ignore
            ]
        )

    @property
    def records(self) -> List[List[Any]]:
        with self._lock:
            return list(self._values)



import copy
from time import sleep

import numpy as np
import pandas as pd

from tune.concepts.space import Rand
from tune.concepts.flow import (
    Trial,
    TrialDecision,
    TrialReport,
    TrialReportHeap,
    TrialReportLogger,
)


def test_trial():
    trial = Trial("abc", {"a": 1}, {"b": 2}, keys=["a", "b"])
    assert "abc" == trial.trial_id
    assert {"a": 1} == trial.params
    assert {"b": 2} == trial.metadata
    assert ["a", "b"] == trial.keys

    t2 = trial.with_params({"c": 3})
    assert "abc" == t2.trial_id
    assert {"c": 3} == t2.params
    assert {"b": 2} == t2.metadata
    assert ["a", "b"] == t2.keys


def test_copy():
    trial = Trial("abc", {"a": 1})
    t1 = trial.with_dfs({"c": pd.DataFrame([[0]])})

    t2 = copy.copy(t1)
    t3 = copy.deepcopy(t1)
    assert trial.trial_id == t2.trial_id == t3.trial_id
    assert t1.dfs is t2.dfs is t3.dfs
    assert 0 == len(trial.dfs)
    assert 1 == len(t1.dfs)
    assert [] == t3.keys


def test_encode_decode():
    p = {"a": 1, "b": Rand(1, 2)}
    trial = Trial("abc", p, {}, keys=["x", "y"], dfs={"v": ""})
    d = trial.jsondict
    assert isinstance(d["params"]["b"], dict)
    t = Trial.from_jsondict(d)
    assert isinstance(t.params["b"], Rand)
    assert ["x", "y"] == t.keys
    assert {} == t.dfs  # dfs will not be serialized


def test_trial_report():
    trial = Trial("abc", {"a": Rand(3, 4)}, {"b": 2})
    report = copy.copy(
        TrialReport(
            trial,
            metric=np.float(0.1),
            params={"c": Rand(1, 2)},
            metadata={"d": 4},
            cost=2.0,
        )
    )
    assert trial is report.trial
    assert 0.1 == report.metric
    assert type(report.metric) == float
    assert {"c": Rand(1, 2)} == report.params
    assert {"d": 4} == report.metadata
    assert 2.0 == report.cost
    assert 0 == report.rung
    assert 0.1 == report.sort_metric

    report = copy.deepcopy(
        TrialReport(trial, metric=np.float(0.111), cost=2.0, rung=4, sort_metric=1.23)
    )
    assert trial is report.trial
    assert 0.111 == report.metric
    assert type(report.metric) == float
    assert {"a": Rand(3, 4)} == report.params
    assert {} == report.metadata
    assert 2.0 == report.cost
    assert 4 == report.rung

    r1 = report.generate_sort_metric(True, 2)
    r2 = report.generate_sort_metric(False, 1)
    r3 = report.with_sort_metric(0.234)
    assert 1.23 == report.sort_metric
    assert 0.11 == r1.sort_metric
    assert -0.1 == r2.sort_metric
    assert 0.234 == r3.sort_metric

    ts = report.log_time
    sleep(0.1)
    report = TrialReport.from_jsondict(report.jsondict)
    nr = report.reset_log_time()
    assert nr.log_time > report.log_time
    assert report.log_time == ts
    assert trial.trial_id == report.trial_id
    assert 0.111 == report.metric
    assert type(report.metric) == float
    assert {"a": Rand(3, 4)} == report.params
    assert {} == report.metadata
    assert 2.0 == report.cost

    assert 3.0 == report.with_cost(3.0).cost
    assert 5 == report.with_rung(5).rung

    td = trial.with_dfs({"a": pd.DataFrame})
    report = TrialReport(td, metric=np.float(0.1))
    assert 0 == len(report.trial.dfs)


def test_trial_report_heap():
    t1 = Trial("a", {})
    r1 = TrialReport(t1, 0.1)
    t2 = Trial("b", {})
    r2 = TrialReport(t2, 0.2)
    t3 = Trial("c", {})
    r3 = TrialReport(t3, 0.3)
    r4 = TrialReport(t3, -0.3)
    h = TrialReportHeap(min_heap=True)
    for r in [r1, r2, r3, r4]:
        h.push(r)
    assert "a" in h
    assert "x" not in h
    for r in [r4, r1, r2]:
        assert h.pop() is r
    assert 0 == len(h)

    h = TrialReportHeap(min_heap=False)
    for r in [r1, r2, r3, r4]:
        h.push(r)
    for r in [r1, r2, r4]:
        assert r in list(h.values())
    for r in [r2, r1, r4]:
        assert h.pop() is r
    assert 0 == len(h)

    # test __lt__, the sort key is sort_metric!
    r5 = TrialReport(t1, metric=0.1, sort_metric=-0.1)
    r6 = TrialReport(t2, metric=0.2, sort_metric=-0.2)
    r7 = TrialReport(t3, metric=0.3, sort_metric=-0.3)

    h = TrialReportHeap(min_heap=True)
    for r in [r7, r6, r5]:
        h.push(r)
    for r in [r7, r6, r5]:
        assert h.pop() is r
    assert 0 == len(h)

    r5 = TrialReport(t1, metric=0.1, cost=0.2, rung=5)
    r6 = TrialReport(t2, metric=0.1, cost=0.3, rung=5)
    r7 = TrialReport(t3, metric=0.1, cost=0.3, rung=6)

    h = TrialReportHeap(min_heap=True)
    for r in [r7, r6, r5]:
        h.push(r)
    for r in [r5, r6, r7]:
        assert h.pop() is r
    assert 0 == len(h)

    # equal case
    r8 = TrialReport(t1, metric=0.1, cost=0.3, rung=6)
    r9 = TrialReport(t2, metric=0.1, cost=0.3, rung=6)

    h = TrialReportHeap(min_heap=False)
    for r in [r8, r9]:
        h.push(r)
    for r in [r8, r9]:
        assert h.pop() is r
    assert 0 == len(h)


def test_trial_decision():
    trial = Trial("abc", {"a": 1}, {"b": Rand(0, 2)})
    report = TrialReport(
        trial, metric=np.float(0.1), params={"c": Rand(0, 3)}, metadata={"d": 4}
    )
    decision = TrialDecision(
        report, budget=0.0, should_checkpoint=True, metadata={"x": 1}, reason="p"
    )
    assert trial is decision.trial
    assert report is decision.report
    assert decision.should_stop
    assert decision.should_checkpoint
    assert {"x": 1} == decision.metadata
    assert "p" == decision.reason
    assert 0.0 == decision.budget

    assert copy.copy(decision) is decision
    assert copy.deepcopy(decision) is decision

    d2 = TrialDecision.from_jsondict(decision.jsondict)
    assert d2.trial_id == trial.trial_id
    assert Rand(0, 3) == d2.report.params["c"]
    assert decision.should_stop
    assert decision.should_checkpoint
    assert {"x": 1} == decision.metadata
    assert "p" == decision.reason

    decision = TrialDecision(
        report, budget=1.0, should_checkpoint=True, metadata={"x": 1}
    )
    assert 1.0 == decision.budget
    assert not decision.should_stop


def test_report_logger():
    class Mock(TrialReportLogger):
        def __init__(self, new_best_only: bool = False):
            super().__init__(new_best_only=new_best_only)
            self._values = []

        def log(self, report: TrialReport) -> None:
            return self._values.append(report)

        @property
        def records(self):
            return self._values

    t1 = Trial("a", dict(a=1, b=2), keys=["x", "y"])
    r1 = TrialReport(t1, 0.8, sort_metric=-0.8)
    t2 = Trial("b", dict(a=11, b=12), keys=["xx", "y"])
    r2 = TrialReport(t2, 0.7, sort_metric=-0.7)
    t3 = Trial("c", dict(a=10, b=20), keys=["x", "y"])
    r3 = TrialReport(t3, 0.9, sort_metric=-0.9)

    b = Mock()
    assert 0 == len(b.records)
    assert b.best is None
    assert b.on_report(r1)
    assert b.on_report(r3)
    assert r3 is b.best
    assert 2 == len(b.records)

    b = Mock()
    assert b.on_report(r3)
    assert not b.on_report(r1)
    assert r3 is b.best
    assert 2 == len(b.records)

    b = Mock(new_best_only=True)
    assert b.on_report(r3)
    assert not b.on_report(r1)
    assert r3 is b.best
    assert 1 == len(b.records)
